@@ -1,11 +1,10 @@
-package httpmock
+package httpmock_test
 
 import (
 	"bytes"
 	"context"
 	"encoding/json"
 	"errors"
-	"io/ioutil"
 	"net"
 	"net/http"
 	"net/url"
@@ -14,36 +13,12 @@ import (
 	"strings"
 	"testing"
 	"time"
+
+	. "github.com/jarcoal/httpmock"
+	"github.com/jarcoal/httpmock/internal"
 )
 
-var testURL = "http://www.example.com/"
-
-func assertBody(t *testing.T, resp *http.Response, expected string) {
-	defer resp.Body.Close()
-
-	data, err := ioutil.ReadAll(resp.Body)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	got := string(data)
-
-	if got != expected {
-		t.Errorf("Expected body: %#v, got %#v", expected, got)
-	}
-}
-
-func TestRouteKey(t *testing.T) {
-	got, expected := noResponder.String(), "NO_RESPONDER"
-	if got != expected {
-		t.Errorf("got: %v, expected: %v", got, expected)
-	}
-
-	got, expected = routeKey{Method: "GET", URL: "/foo"}.String(), "GET /foo"
-	if got != expected {
-		t.Errorf("got: %v, expected: %v", got, expected)
-	}
-}
+const testURL = "http://www.example.com/"
 
 func TestMockTransport(t *testing.T) {
 	Activate()
@@ -53,26 +28,68 @@ func TestMockTransport(t *testing.T) {
 	body := `["hello world"]` + "\n"
 
 	RegisterResponder("GET", url, NewStringResponder(200, body))
+	RegisterResponder("GET", `=~/xxx\z`, NewStringResponder(200, body))
 
-	// Read it as a simple string (ioutil.ReadAll will trigger io.EOF)
+	// Read it as a simple string (ioutil.ReadAll of assertBody will
+	// trigger io.EOF)
 	func() {
 		resp, err := http.Get(url)
 		if err != nil {
 			t.Fatal(err)
 		}
-		defer resp.Body.Close()
-
-		data, err := ioutil.ReadAll(resp.Body)
-		if err != nil {
-			t.Fatal(err)
-		}
-
-		if string(data) != body {
+		if !assertBody(t, resp, body) {
 			t.FailNow()
 		}
 
 		// the http client wraps our NoResponderFound error, so we just try and match on text
-		if _, err := http.Get(testURL); !strings.Contains(err.Error(), NoResponderFound.Error()) {
+		_, err = http.Get(testURL)
+		if err == nil {
+			t.Fatal("An error should occur")
+		}
+		if !strings.HasSuffix(err.Error(), NoResponderFound.Error()) {
+			t.Fatal(err)
+		}
+
+		// Use wrongly cased method, the error should warn us
+		req, err := http.NewRequest("Get", url, nil)
+		if err != nil {
+			t.Fatal(err)
+		}
+		c := http.Client{}
+		_, err = c.Do(req)
+		if err == nil {
+			t.Fatal("An error should occur")
+		}
+		if !strings.HasSuffix(err.Error(),
+			NoResponderFound.Error()+" for method Get, but one matches method GET") {
+			t.Fatal(err)
+		}
+
+		// Use POST instead of GET, the error should warn us
+		req, err = http.NewRequest("POST", url, nil)
+		if err != nil {
+			t.Fatal(err)
+		}
+		_, err = c.Do(req)
+		if err == nil {
+			t.Fatal("An error should occur")
+		}
+		if !strings.HasSuffix(err.Error(),
+			NoResponderFound.Error()+" for method POST, but one matches method GET") {
+			t.Fatal(err)
+		}
+
+		// Same using a regexp responder
+		req, err = http.NewRequest("POST", "http://pipo.com/xxx", nil)
+		if err != nil {
+			t.Fatal(err)
+		}
+		_, err = c.Do(req)
+		if err == nil {
+			t.Fatal("An error should occur")
+		}
+		if !strings.HasSuffix(err.Error(),
+			NoResponderFound.Error()+" for method POST, but one matches method GET") {
 			t.Fatal(err)
 		}
 	}()
@@ -125,34 +142,26 @@ func TestMockTransportDefaultMethod(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	defer resp.Body.Close()
 
-	data, err := ioutil.ReadAll(resp.Body)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	if string(data) != body {
-		t.FailNow()
-	}
+	assertBody(t, resp, body)
 }
 
 func TestMockTransportReset(t *testing.T) {
 	DeactivateAndReset()
 
-	if len(DefaultTransport.responders) > 0 {
+	if DefaultTransport.NumResponders() > 0 {
 		t.Fatal("expected no responders at this point")
 	}
 
-	RegisterResponder("GET", testURL, nil)
+	RegisterResponder("GET", testURL, NewStringResponder(200, "hey"))
 
-	if len(DefaultTransport.responders) != 1 {
+	if DefaultTransport.NumResponders() != 1 {
 		t.Fatal("expected one responder")
 	}
 
 	Reset()
 
-	if len(DefaultTransport.responders) > 0 {
+	if DefaultTransport.NumResponders() > 0 {
 		t.Fatal("expected no responders as they were just reset")
 	}
 }
@@ -163,10 +172,6 @@ func TestMockTransportNoResponder(t *testing.T) {
 
 	Reset()
 
-	if DefaultTransport.noResponder != nil {
-		t.Fatal("expected noResponder to be nil")
-	}
-
 	if _, err := http.Get(testURL); err == nil {
 		t.Fatal("expected to receive a connection error due to lack of responders")
 	}
@@ -177,14 +182,26 @@ func TestMockTransportNoResponder(t *testing.T) {
 	if err != nil {
 		t.Fatal("expected request to succeed")
 	}
+	assertBody(t, resp, "hello world")
 
-	data, err := ioutil.ReadAll(resp.Body)
-	if err != nil {
-		t.Fatal(err)
+	// Using NewNotFoundResponder()
+	RegisterNoResponder(NewNotFoundResponder(nil))
+	_, err = http.Get(testURL)
+	if err == nil {
+		t.Fatal("an error should occur")
+	}
+	if !strings.HasSuffix(err.Error(), "Responder not found for GET http://www.example.com/") {
+		t.Fatalf("Unexpected error content: %s", err)
 	}
 
-	if string(data) != "hello world" {
-		t.Fatal("expected body to be 'hello world'")
+	// Help the user in case a Responder exists for another method
+	RegisterResponder("POST", testURL, NewStringResponder(200, "hello world"))
+	_, err = http.Get(testURL)
+	if err == nil {
+		t.Fatal("an error should occur")
+	}
+	if !strings.HasSuffix(err.Error(), "Responder not found for GET http://www.example.com/, but one matches method POST") {
+		t.Fatalf("Unexpected error content: %s", err)
 	}
 }
 
@@ -197,6 +214,7 @@ func TestMockTransportQuerystringFallback(t *testing.T) {
 
 	for _, suffix := range []string{"?", "?hello=world", "?hello=world#foo", "?hello=world&hello=all", "#foo"} {
 		reqURL := testURL + suffix
+		t.Log(reqURL)
 
 		// make a request for the testURL with a querystring
 		resp, err := http.Get(reqURL)
@@ -204,14 +222,7 @@ func TestMockTransportQuerystringFallback(t *testing.T) {
 			t.Fatalf("expected request %s to succeed", reqURL)
 		}
 
-		data, err := ioutil.ReadAll(resp.Body)
-		if err != nil {
-			t.Fatalf("%s error: %s", reqURL, err)
-		}
-
-		if string(data) != "hello world" {
-			t.Fatalf("expected body of %s to be 'hello world'", reqURL)
-		}
+		assertBody(t, resp, "hello world")
 	}
 }
 
@@ -357,20 +368,16 @@ func TestMockTransportPathOnlyFallback(t *testing.T) {
 		RegisterResponder("GET", test.Responder, NewStringResponder(200, "hello world"))
 
 		for _, reqURL := range test.Paths {
+			t.Logf("%s: %s", test.Responder, reqURL)
+
 			// make a request for the testURL with a querystring
 			resp, err := http.Get(reqURL)
 			if err != nil {
-				t.Fatalf("%s: expected request %s to succeed", test.Responder, reqURL)
+				t.Errorf("%s: expected request %s to succeed", test.Responder, reqURL)
+				continue
 			}
 
-			data, err := ioutil.ReadAll(resp.Body)
-			if err != nil {
-				t.Fatalf("%s: %s error: %s", test.Responder, reqURL, err)
-			}
-
-			if string(data) != "hello world" {
-				t.Fatalf("%s: expected body of %s to be 'hello world'", test.Responder, reqURL)
-			}
+			assertBody(t, resp, "hello world")
 		}
 
 		DeactivateAndReset()
@@ -433,16 +440,7 @@ func TestMockTransportNonDefault(t *testing.T) {
 		t.Fatal(err)
 	}
 
-	defer resp.Body.Close()
-
-	data, err := ioutil.ReadAll(resp.Body)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	if string(data) != body {
-		t.FailNow()
-	}
+	assertBody(t, resp, body)
 }
 
 func TestMockTransportRespectsCancel(t *testing.T) {
@@ -689,9 +687,25 @@ func TestMockTransportCallCountZero(t *testing.T) {
 	if !reflect.DeepEqual(info, expectedInfo) {
 		t.Fatalf("did not correctly reset the call count info. expected it to be \n %+v\n but it was \n %+v", expectedInfo, info)
 	}
+
+	// Unregister each responder
+	RegisterResponder("GET", url, nil)
+	RegisterResponder("POST", "=~gitlab", nil)
+
+	info = GetCallCountInfo()
+	expectedInfo = map[string]int{
+		// this one remains as it is not directly related to a registered
+		// responder but a consequence of a regexp match
+		"POST " + url2: 0,
+	}
+	if !reflect.DeepEqual(info, expectedInfo) {
+		t.Fatalf("did not correctly reset the call count info. expected it to be \n %+v\n but it was \n %+v", expectedInfo, info)
+	}
 }
 
 func TestRegisterResponderWithQuery(t *testing.T) {
+	Reset()
+
 	// Just in case a panic occurs
 	defer DeactivateAndReset()
 
@@ -755,21 +769,46 @@ func TestRegisterResponderWithQuery(t *testing.T) {
 			RegisterResponderWithQuery("GET", testURLPath, query, NewStringResponder(200, body))
 
 			for _, url := range test.URLs {
+				t.Logf("query=%v URL=%s", query, url)
+
 				req, err := http.NewRequest("GET", url, nil)
 				if err != nil {
 					t.Fatal(err)
 				}
+
 				resp, err := client.Do(req)
 				if err != nil {
 					t.Fatal(err)
 				}
-				data, err := ioutil.ReadAll(resp.Body)
-				resp.Body.Close()
+
+				assertBody(t, resp, body)
+			}
+
+			if info := GetCallCountInfo(); len(info) != 1 {
+				t.Fatalf("%s: len(GetCallCountInfo()) should be 1 but contains %+v", testURLPath, info)
+			}
+
+			// Remove...
+			RegisterResponderWithQuery("GET", testURLPath, query, nil)
+			if info := GetCallCountInfo(); len(info) != 0 {
+				t.Fatalf("did not correctly reset the call count info, it still contains %+v", info)
+			}
+
+			for _, url := range test.URLs {
+				t.Logf("query=%v URL=%s", query, url)
+
+				req, err := http.NewRequest("GET", url, nil)
 				if err != nil {
 					t.Fatal(err)
 				}
-				if string(data) != body {
-					t.Fatalf("query=%v URL=%s: %s ≠ %s", query, url, string(data), body)
+
+				_, err = client.Do(req)
+				if err == nil {
+					t.Fatalf("No error occurred for %s", url)
+				}
+
+				if !strings.HasSuffix(err.Error(), "no responder found") {
+					t.Errorf("Not expected error suffix: %s", err)
 				}
 			}
 
@@ -802,27 +841,18 @@ func TestRegisterResponderWithQueryPanic(t *testing.T) {
 			PanicPrefix: `path begins with "=~", RegisterResponder should be used instead of RegisterResponderWithQuery`,
 		},
 	} {
-		var (
-			didntPanic bool
-			panicVal   interface{}
-		)
-		func() {
-			defer func() {
-				panicVal = recover()
-			}()
-
+		panicked, panicStr := catchPanic(func() {
 			RegisterResponderWithQuery("GET", test.Path, test.Query, resp)
-			didntPanic = true
-		}()
+		})
 
-		if didntPanic {
-			t.Fatalf("RegisterResponderWithQuery + query=%v did not panic", test.Query)
+		if !panicked {
+			t.Errorf("RegisterResponderWithQuery + query=%v did not panic", test.Query)
+			continue
 		}
 
-		panicStr, ok := panicVal.(string)
-		if !ok || !strings.HasPrefix(panicStr, test.PanicPrefix) {
+		if !strings.HasPrefix(panicStr, test.PanicPrefix) {
 			t.Fatalf(`RegisterResponderWithQuery + query=%v panic="%v" expected prefix="%v"`,
-				test.Query, panicVal, test.PanicPrefix)
+				test.Query, panicStr, test.PanicPrefix)
 		}
 	}
 }
@@ -842,14 +872,7 @@ func TestRegisterRegexpResponder(t *testing.T) {
 		t.Fatalf("expected request %s to succeed", testURL)
 	}
 
-	data, err := ioutil.ReadAll(resp.Body)
-	if err != nil {
-		t.Fatalf("%s error: %s", testURL, err)
-	}
-
-	if string(data) != "second" {
-		t.Fatalf("expected body of %s to be 'hello world'", testURL)
-	}
+	assertBody(t, resp, "second")
 }
 
 func TestSubmatches(t *testing.T) {
@@ -858,24 +881,7 @@ func TestSubmatches(t *testing.T) {
 		t.Fatal(err)
 	}
 
-	var req2 *http.Request
-
-	t.Run("setSubmatches", func(t *testing.T) {
-		req2 = setSubmatches(req, nil)
-		if req2 != req {
-			t.Error("setSubmatches(req, nil) should return the same request")
-		}
-
-		req2 = setSubmatches(req, []string{})
-		if req2 != req {
-			t.Error("setSubmatches(req, []string{}) should return the same request")
-		}
-
-		req2 = setSubmatches(req, []string{"foo", "123", "-123", "12.3"})
-		if req2 == req {
-			t.Error("setSubmatches(req, []string{...}) should NOT return the same request")
-		}
-	})
+	req2 := internal.SetSubmatches(req, []string{"foo", "123", "-123", "12.3"})
 
 	t.Run("GetSubmatch", func(t *testing.T) {
 		_, err := GetSubmatch(req, 1)
@@ -1081,78 +1087,12 @@ func TestSubmatches(t *testing.T) {
 }
 
 func TestCheckStackTracer(t *testing.T) {
-	req, err := http.NewRequest("GET", "http://foo.bar/", nil)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	// no error
-	gotErr := checkStackTracer(req, nil)
-	if gotErr != nil {
-		t.Errorf(`checkStackTracer(nil) should return nil, not %v`, gotErr)
-	}
-
-	// Classic error
-	err = errors.New("error")
-	gotErr = checkStackTracer(req, err)
-	if err != gotErr {
-		t.Errorf(`checkStackTracer(err) should return %v, not %v`, err, gotErr)
-	}
-
-	// stackTracer without customFn
-	origErr := errors.New("foo")
-	errTracer := stackTracer{
-		err: origErr,
-	}
-	gotErr = checkStackTracer(req, errTracer)
-	if gotErr != origErr {
-		t.Errorf(`Returned error mismatch, expected: %v, got: %v`, origErr, gotErr)
-	}
-
-	// stackTracer with nil error & without customFn
-	errTracer = stackTracer{}
-	gotErr = checkStackTracer(req, errTracer)
-	if gotErr != nil {
-		t.Errorf(`Returned error mismatch, expected: nil, got: %v`, gotErr)
-	}
-
-	// stackTracer
-	var mesg string
-	errTracer = stackTracer{
-		err: origErr,
-		customFn: func(args ...interface{}) {
-			mesg = args[0].(string)
-		},
-	}
-	gotErr = checkStackTracer(req, errTracer)
-	if !strings.HasPrefix(mesg, "foo\nCalled from ") || strings.HasSuffix(mesg, "\n") {
-		t.Errorf(`mesg does not match "^foo\nCalled from .*[^\n]\z", it is "` + mesg + `"`)
-	}
-	if gotErr != origErr {
-		t.Errorf(`Returned error mismatch, expected: %v, got: %v`, origErr, gotErr)
-	}
-
-	// stackTracer with nil error but customFn
-	mesg = ""
-	errTracer = stackTracer{
-		customFn: func(args ...interface{}) {
-			mesg = args[0].(string)
-		},
-	}
-	gotErr = checkStackTracer(req, errTracer)
-	if !strings.HasPrefix(mesg, "GET http://foo.bar/\nCalled from ") || strings.HasSuffix(mesg, "\n") {
-		t.Errorf(`mesg does not match "^foo\nCalled from .*[^\n]\z", it is "` + mesg + `"`)
-	}
-	if gotErr != nil {
-		t.Errorf(`Returned error mismatch, expected: nil, got: %v`, gotErr)
-	}
-
 	// Full test using Trace() Responder
 	Activate()
 	defer Deactivate()
 
 	const url = "https://foo.bar/"
-	mesg = ""
+	var mesg string
 	RegisterResponder("GET", url,
 		NewStringResponder(200, "{}").
 			Trace(func(args ...interface{}) { mesg = args[0].(string) }))
@@ -1161,14 +1101,8 @@ func TestCheckStackTracer(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	defer resp.Body.Close()
 
-	data, err := ioutil.ReadAll(resp.Body)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	if string(data) != "{}" {
+	if !assertBody(t, resp, "{}") {
 		t.FailNow()
 	}
 
@@ -1178,3 +1112,64 @@ func TestCheckStackTracer(t *testing.T) {
 		t.Errorf("Bad mesg: <%v>", mesg)
 	}
 }
+
+func TestCheckMethod(t *testing.T) {
+	mt := NewMockTransport()
+
+	var (
+		panicked bool
+		panicStr string
+	)
+
+	//
+	// Panics
+	checkPanic := func() {
+		helper(t).Helper()
+		if panicStr != `You probably want to use method "GET" instead of "get"? If not and so want to disable this check, set MockTransport.DontCheckMethod field to true` {
+			if panicked {
+				t.Errorf("Wrong panic mesg: %s", panicStr)
+			} else {
+				t.Error("Did not panic!")
+			}
+		}
+	}
+
+	panicked, panicStr = catchPanic(func() {
+		mt.RegisterResponder("get", "/pipo", NewStringResponder(200, ""))
+	})
+	checkPanic()
+
+	panicked, panicStr = catchPanic(func() {
+		mt.RegisterRegexpResponder("get", regexp.MustCompile("."), NewStringResponder(200, ""))
+	})
+	checkPanic()
+
+	panicked, panicStr = catchPanic(func() {
+		mt.RegisterResponderWithQuery("get", "/pipo", url.Values(nil), NewStringResponder(200, ""))
+	})
+	checkPanic()
+
+	//
+	// No longer panics
+	checkNoPanic := func() {
+		helper(t).Helper()
+		if panicked {
+			t.Errorf("Should not panic! but %s", panicStr)
+		}
+	}
+	mt.DontCheckMethod = true
+	panicked, panicStr = catchPanic(func() {
+		mt.RegisterResponder("get", "/pipo", NewStringResponder(200, ""))
+	})
+	checkNoPanic()
+
+	panicked, panicStr = catchPanic(func() {
+		mt.RegisterRegexpResponder("get", regexp.MustCompile("."), NewStringResponder(200, ""))
+	})
+	checkNoPanic()
+
+	panicked, panicStr = catchPanic(func() {
+		mt.RegisterResponderWithQuery("get", "/pipo", url.Values(nil), NewStringResponder(200, ""))
+	})
+	checkNoPanic()
+}
